
import atexit
import os
import subprocess
import time
import threading

from django.apps import AppConfig
from django.conf import settings


# time in seconds to check for the worker:
try:
    WORKER_MONITOR_INTERVALL = settings.AUTOTASK_WORKER_MONITOR_INTERVALL
except AttributeError:
    WORKER_MONITOR_INTERVALL = 5

# path to the executable:
try:
    WORKER_EXECUTABLE = settings.AUTOTASK_WORKER_EXECUTABLE
except AttributeError:
    WORKER_EXECUTABLE = 'python'


class ExitHandler:

    def __init__(self, process):
        self.process = process

    def __call__(self):
        self.process.terminate()
        self.delete_periodic_tasks()

    def delete_periodic_tasks(self):
        """
        Tasks are persistent in the db. Periodic tasks are read in at
        process start and will not expire. So they should get deleted
        here.
        """
        from .models import TaskQueue  # noqa
        qs = TaskQueue.objects.filter(is_periodic=True)
        if qs.count():
            qs.delete()


class Supervisor:
    """
    Supervisor for the worker process.
    Starts and monitors the worker.
    """
    def __init__(self):
        self.exit_handler = None

    def __call__(self):
        process = self.start_worker()
        self.set_exit_handler(process)
        self.monitor_worker(process)

    def monitor_worker(self, process):
        while True:
            return_code = process.poll()
            if return_code:
                # bad things happend
                process = self.restart_process()
            else:
                time.sleep(WORKER_MONITOR_INTERVALL)

    def restart_process(self):
        atexit.unregister(self.exit_handler)
        process = self.start_worker()
        self.set_exit_handler(process)
        return process

    def set_exit_handler(self, process):
        self.exit_handler = ExitHandler(process)
        atexit.register(self.exit_handler)

    def start_worker(self):
        os.environ.setdefault('DJANGO_AUTOTASK', 'true')
        process = subprocess.Popen(
            [WORKER_EXECUTABLE, 'manage.py', 'run_autotask'],
            env=os.environ,
            cwd=settings.BASE_DIR)
        return process


class AutotaskConfig(AppConfig):
    name = 'autotask'
    is_called = False

    def ready(self):
        if self.is_called:
            return
        self.is_called = True
        if not os.environ.get('DJANGO_AUTOTASK'):
            self.start_supervisor()

    def start_supervisor(self):
        """
        Start supervisor as daemon thread so the supervisor-thread will
        not block on shut-down.
        """
        supervisor = Supervisor()
        thread = threading.Thread(target=supervisor, daemon=True)
        thread.start()



# -*- coding: utf-8 -*-
# Generated by Django 1.9.4 on 2016-04-01 18:51
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='TaskQueue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('scheduled', models.DateTimeField(verbose_name='scheduled')),
                ('module', models.CharField(max_length=256, verbose_name='Module Name')),
                ('function', models.CharField(max_length=256, verbose_name='Function Name')),
                ('arguments', models.BinaryField(blank=True, verbose_name='Arguments')),
                ('is_periodic', models.BooleanField(default=False, verbose_name='Is periodic task')),
                ('timedelta', models.DurationField(blank=True, null=True, verbose_name='timedelta')),
                ('cron_data', models.BinaryField(blank=True, verbose_name='cron data')),
                ('status', models.IntegerField(choices=[(1, 'waiting'), (2, 'running'), (3, 'done'), (4, 'error')], default=1, verbose_name='Status')),
                ('retries', models.IntegerField(default=0, verbose_name='Retries')),
                ('result', models.BinaryField(blank=True, verbose_name='Result')),
                ('ttl', models.DurationField(blank=True, null=True, verbose_name='time to live')),
                ('expire', models.DateTimeField(blank=True, null=True, verbose_name='expire')),
                ('error_message', models.TextField(blank=True, verbose_name='error message')),
            ],
        ),
    ]



"""
Entry-point for starting the workers.
"""

from django.core.management.base import BaseCommand
from autotask.workers import start_workers


class Command(BaseCommand):
    help = 'starts the autotask workers'

    def handle(self, *args, **options):
        start_workers()




import calendar
import datetime
import re

from django.conf import settings
from django.utils.timezone import (
    localtime,
    make_aware,
    now,
)


class CronScheduler:
    """
    Schedules a cron task.

    Usage:

    cs = CronScheduler()
    next_schedule = cs.get_next_schedule()
    """

    def __init__(self, last_schedule=None, minutes=None, hours=None,
                 dow=None, months=None, dom=None, crontab=None):
        """
        Inits the scheduler with values according the crontab-format.

        All arguments are optional. If no arguments are given the task
        is scheduled from now on every minute.

        last_schedule: last time the according tasks has run
        minutes: list of integers when a task should run in an hour [0-59]
        hours: list of integers when a task should run during a day [0-23]
        dow: list of integers for the days of week when a task should
        run [0-6] (0 for monday up to 6 for sunday).
        month: list of integers for the months a task should run [1-12]
        dom: list of intergs for the day in a month a task should run [1-31]
        """
        self.date = last_schedule or now()
        if not self.parse_crontab(crontab):
            self.minutes = minutes
            self.hours = hours
            self.dow = dow
            self.months = months
            self.dom = dom

    def parse_crontab(self, crontab):
        """
        Parses a simple crontab-string with five patterns:

        Minute [0,59]
        Hour [0,23]
        Day of the month [1,31]
        Month of the year [1,12]
        Day of the week ([0,6] with 0=Sunday)

        Each of these patterns can be either an asterisk (meaning all
        valid values), an element, or a list of elements separated by
        commas. An element shall be either a number or two numbers
        separated by a hyphen (meaning an inclusive range).

        A day-range for a month like 1-4,12,20-25 is also ok. But keep
        in mind that no spaces are allowed as spaces are the separators
        between the patterns.

        Some examples:

        * * * * *               runs every minute
        5 * * * *               runs every five minutes
        30 7 10-15 4,7 0        runs at 7:30 on mondays and also from the
                                10th to 15th of a month, but only in april
                                and july

        Returns a boolean whether the crontab could be parsed (with the
        content *not* checked for correctness).
        """
        if not crontab:
            return False
        outer = []
        for item in crontab.split():
            item = item.strip()
            if item == '*':
                outer.append(None)
                continue
            inner = []
            for element in item.split(','):
                mo = re.match(r'(\d+)-(\d+)', element)
                if mo:
                    inner.extend([i for i in range(
                        int(mo.group(1)), int(mo.group(2))+1)])
                else:
                    inner.append(int(element))
            outer.append(inner)
        if len(outer) == 5:
            self.minutes, self.hours, self.dom, self.months, self.dow = outer
            return True
        return False

    def get_next_schedule(self):
        """
        Returns the next schedule based on the current date as a
        datetime-object.
        """
        self.date = localtime(self.date)
        next_minute = self.get_next_minute()
        if next_minute:
            if self.hour_is_allowed():
                if self.month_is_allowed():
                    return self._get_schedule(minute=next_minute)
        return self.get_schedule_next_hour()

    def get_schedule_next_hour(self):
        """
        Returns the next schedule as a datetime-object.
        The first minute is valid, the next hour has to be found.
        """
        first_minute = self.get_first_minute()
        next_hour = self.get_next_hour()
        if next_hour:
            if self.day_is_allowed():
                if self.month_is_allowed():
                    schedule = self._get_schedule(
                        hour=next_hour, minute=first_minute)
                else:
                    first_hour = self.get_first_hour()
                    schedule = self.get_schedule_next_month(
                        hour=first_hour, minute=first_minute)
                return schedule
        return self.get_schedule_next_day(first_minute)

    def get_schedule_next_day(self, minute):
        """
        Returns the next schedule as a datetime-object.
        The first minute is valid, the next hour has to be found.
        """
        first_hour = self.get_first_hour()
        next_day = self.get_next_day()
        if next_day:
            if self.month_is_allowed():
                return self._get_schedule(
                    day=next_day, hour=first_hour, minute=minute)
        return self.get_schedule_next_month(first_hour, minute)

    def get_schedule_next_month(self, hour, minute):
        """
        Returns the next schedule as a datetime-object.
        The first minute and hour are valid, the next day has to be
        found.
        """
        year, month = self.get_next_month()
        day = self.get_first_day(year, month)
        schedule = self._get_schedule(
            year=year, month=month, day=day, hour=hour, minute=minute)
        return schedule

    def hour_is_allowed(self):
        """
        Returns a boolean whether the current hour is allowed for the
        current date.
        """
        if self.hours:
            return self.date.hour in self.hours
        # else: every hour is allowed
        return True

    def day_is_allowed(self):
        """
        Returns a boolean whether the current day is allowed for the
        current date.
        """
        if not self.dow and not self.dom:
            # every day is alloed:
            return True
        if self.date.day in self.dom:
            return True
        # now check whether day of week is allowed
        this = self.date
        weekday = calendar.weekday(this.year, this.month, this.day)
        return weekday in self.dow

    def month_is_allowed(self):
        """
        Returns a boolean whether the current month is allowed for the
        current date.
        """
        if self.months:
            return self.date.month in self.months
        # else: every month is allowed
        return True

    def get_first_minute(self):
        """Returns the first allowed minute of an hour."""
        return self.minutes[0] if self.minutes else 0

    def get_first_hour(self):
        """Returns the first allowed hour of a day."""
        return self.hours[0] if self.hours else 0

    def get_first_day(self, year, month):
        """Returns the first allowed day of the given month and year."""
        if not (self.dow or self.dom):
            # every day is allowed
            return 1
        first_dom_day = self.dom[0] if self.dom else None
        first_dow_day = None
        if self.dow:
            first_weekday, _ = calendar.monthrange(year, month)
            if first_weekday in self.dow:
                return 1
            first_allowed_weekday = get_next_value(first_weekday, self.dow)
            if first_allowed_weekday:
                delta = first_allowed_weekday - first_weekday
            else:
                delta = first_weekday - self.dow[0]
                if delta > 0:
                    delta = 7 - delta
            first_dow_day = 1 + delta
        if first_dow_day and first_dom_day:
            return min(first_dom_day, first_dow_day)
        return first_dom_day or first_dow_day

    def get_next_minute(self):
        """
        Returns next allowed minute according to self.date.
        Returns None if no next allowed minute can be found.
        """
        if self.minutes:
            return get_next_value(self.date.minute, self.minutes)
        return self.date.minute + 1 if self.date.minute < 59 else None

    def get_next_hour(self):
        """
        Returns next allowed hour according to self.date.
        Returns None if no next allowed hour can be found.
        """
        if self.hours:
            return get_next_value(self.date.hour, self.hours)
        return self.date.hour + 1 if self.date.hour < 23 else None

    def get_next_day(self):
        """
        Returns next allowed day according to self.date.
        Returns None if no next allowed day can be found.
        """
        _, max_days = calendar.monthrange(
            self.date.year, self.date.month)
        if not self.dow and not self.dom:
            next_day = self.date.day + 1
        else:
            next_dom_day = None
            next_dow_day = None
            if self.dom:
                next_dom_day = get_next_value(self.date.day, self.dom)
            if self.dow:
                weekday = self.date.weekday()
                next_weekday = get_next_value(weekday, self.dow) or self.dow[0]
                delta_days = next_weekday - weekday
                if delta_days <= 0:
                    delta_days += 7
                next_dow_day = self.date.day + delta_days
            if next_dom_day and next_dow_day:
                next_day = min(next_dom_day, next_dow_day)
            else:
                next_day = next_dom_day or next_dow_day
        if next_day and next_day > max_days:
            next_day = None
        return next_day

    def get_next_month(self):
        """
        Returns next allowed tuple of (month, year) according to
        self.date.
        """
        year = self.date.year
        if self.months:
            # just selective months are allowed
            next_month = get_next_value(self.date.month, self.months)
            if next_month is None:
                year += 1
                next_month = get_next_value(0, self.months)
        else:
            # every month is allowed
            next_month = self.date.month + 1
            if next_month > 12:
                next_month = 1
                year += 1
        return year, next_month

    def _get_schedule(self, year=None, month=None,
                      day=None, hour=None, minute=None):
        """
        Returns a datetime-object with the arguments or the values from
        self.date as default.
        """
        year = year or self.date.year
        month = month or self.date.month
        day = day or self.date.day
        # hour and minute are allowed to be 0, therefore test for None
        hour = hour if hour is not None else self.date.hour
        minute = minute if minute is not None else self.date.minute
        schedule = datetime.datetime(year, month, day, hour, minute)
        if settings.USE_TZ:
            schedule = make_aware(schedule)
        return schedule


def get_next_value(start_value, values):
    """
    Returns next value from a sequence of values if there is a value >
    start_value. Returns None if no such value is found.
    """
    if values is not None:
        for value in values:
            if value > start_value:
                return value
    return None




import datetime
import time
import pytest

from django.utils.timezone import now

from autotask.models import (
    WAITING,
    RUNNING,
    DONE,
    ERROR,
    TaskQueue,
)
from autotask.tasks import (
    DelayedTask,
    delayed_task,
    periodic_task,
    cron_task,
)
from autotask.workers import (
    clean_queue,
    TaskHandler,
)


# test functions
def add(a, b):
    return a + b

@delayed_task()
def add2(a, b):
    return a + b

@delayed_task(retries=1)
def add3(a, b):
    return a + b

@delayed_task(ttl=0.1)
def add4(a, b):
    return a + b

@delayed_task()
def mult(a, b):
    return a * b

@delayed_task(delay=0.01)
def mult2(a, b):
    return a * b

@periodic_task(seconds=0.02, start_now=True)
def gettime():
    return time.time()

@periodic_task(seconds=0.02, start_now=False)
def gettime2():
    return time.time()


@pytest.mark.django_db
class TestAutotask(object):
    @pytest.fixture(autouse=True)
    def set_up(self):
        pass

    def test_delayed_task_01(self):
        """Test whether tasks get stored."""
        n = TaskQueue.objects.all().count()
        assert 0 == n
        dt = delayed_task()
        f = dt(add)
        f(2, 2)
        n = TaskQueue.objects.all().count()
        assert 1 == n

    def test_delayed_task_02(self):
        """Test for return object."""
        dt = delayed_task()
        f = dt(add)
        r = f(2, 2)
        assert True is isinstance(r, DelayedTask)

    def test_decorator_01(self):
        """Test for return object."""
        r = add2(2, 2)
        assert True is isinstance(r, DelayedTask)

    def test_decorator_02(self):
        """Test for return attributes."""
        r = add2(2, 2)
        assert r.ready is False
        assert r.status == WAITING

    def test_taskhandler_01(self):
        """should return None if no task found."""
        th = TaskHandler()
        assert th.get_next_task() is None

    def test_taskhandler_02(self):
        """Test whether is an entry in the database."""
        r = add2(2, 2)
        n = TaskQueue.objects.all().count()
        assert 1 == n

    def test_taskhandler_03(self):
        """Test for accessing this entry."""
        r = add2(2, 2)
        th = TaskHandler()
        task = th.get_next_task()
        assert task.status == RUNNING

    def test_taskhandler_04(self):
        """Test for accessing the first entry."""
        r = add2(2, 2)
        p = mult(2, 2)
        th = TaskHandler()
        task = th.get_next_task()
        assert task.function == 'add2_delayed'

    @pytest.mark.parametrize(
        'a, b, result, status, ready', [
            (2, 3, 5, DONE, True),
            (2, "a", None, ERROR, True),
            ('a', 'b', 'ab', DONE, True),
        ])
    def test_taskhandler_05(self, a, b, result, status, ready):
        """test error handling."""
        r = add2(a, b)
        th = TaskHandler()
        task = th.get_next_task()
        th.handle_task(task)
        assert task.status == status
        assert r.status == status
        assert r.result == result
        assert r.ready is True

    @pytest.mark.parametrize(
        'delay, ready', [
            (0, False),
            (0.012, True),
        ])
    def test_taskhandler_06(self, delay, ready):
        """test delay."""
        th = TaskHandler()
        r = mult2(2, 2)
        time.sleep(delay)
        task = th.get_next_task()
        if task:
            th.handle_task(task)
        assert r.ready is ready

    def test_taskhandler_07(self):
        """test retries."""
        r = add3(2, 'c')
        th = TaskHandler()
        task = th.get_next_task()
        th.handle_task(task)
        assert r.status == WAITING
        # should be false because the task is rescheduled
        assert r.ready is False
        time.sleep(1)
        task = th.get_next_task()
        # should be None becaus schedule delay is 1 sec.
        assert task is None
        time.sleep(1.5)
        task = th.get_next_task()
        assert task is not None
        th.handle_task(task)
        # now status should be ERROR and task is done
        assert r.status == ERROR
        assert r.ready is True

    @pytest.mark.parametrize(
        'ttl, result', [
            (0, 5),
            (0.05, 5),
            (0.15, None),
        ])
    def test_taskhandler_08(self, ttl, result):
        """test ttl."""
        r = add4(2, 3)
        th = TaskHandler()
        task = th.get_next_task()
        th.handle_task(task)
        time.sleep(ttl)
        clean_queue()
        assert r.result == result

    def test_periodic_task_01(self):
        """test periodic execution."""
        r = gettime()
        th = TaskHandler()
        # run task and store the result
        task = th.get_next_task()
        th.handle_task(task)
        result = r.result
        # wait shorter than timedelta and try to get the task again.
        time.sleep(0.01)
        task = th.get_next_task()
        assert task is None
        # wait timedelta and run again
        # result should have changed.
        time.sleep(0.02)
        task = th.get_next_task()
        if task:
            th.handle_task(task)
        assert result != r.result

    def test_periodic_task_02(self):
        """test delayed periodic execution."""
        r = gettime2()
        th = TaskHandler()
        # try to get the task
        task = th.get_next_task()
        assert task is None
        # wait shorter than timedelta and try again.
        # result should not change.
        time.sleep(0.01)
        task = th.get_next_task()
        assert task is None
        # now the task should be available.
        time.sleep(0.02)
        task = th.get_next_task()
        assert task is not None




from datetime import timedelta

from django.db import models
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _


WAITING = 1
RUNNING = 2
DONE = 3
ERROR = 4

STATUS_CHOICES = (
    (WAITING, 'waiting'),
    (RUNNING, 'running'),
    (DONE, 'done'),
    (ERROR, 'error'),
)


class TaskQueue(models.Model):

    scheduled = models.DateTimeField(
        _('scheduled'))

    module = models.CharField(
        _('Module Name'),
        max_length=256)

    function = models.CharField(
        _('Function Name'),
        max_length=256)

    arguments = models.BinaryField(
        _('Arguments'),
        blank=True)

    is_periodic = models.BooleanField(
        _('Is periodic task'),
        default=False)

    timedelta = models.DurationField(
        _('timedelta'),
        blank=True,
        null=True)

    cron_data = models.BinaryField(
        _('cron data'),
        blank=True)

    status = models.IntegerField(
        _('Status'),
        choices=STATUS_CHOICES,
        default=WAITING)

    retries = models.IntegerField(
        _('Retries'),
        default=0)

    result = models.BinaryField(
        _('Result'),
        blank=True)

    ttl = models.DurationField(
        _('time to live'),
        blank=True,
        null=True)

    expire = models.DateTimeField(
        _('expire'),
        blank=True,
        null=True)

    error_message = models.TextField(
        _('error message'),
        blank=True)

    def __str__(self):
        task = 'cron' if self.is_periodic else 'task'
        info = '{}: {}'.format(task, self.function)
        return info

    def save(self, **kwargs):
        if not self.ttl:
            self.ttl = timedelta()
        if not self.timedelta:
            # prevent an uninitialized task from running mad
            self.timedelta = timedelta(seconds=3600)
        if not self.scheduled:
            # try to avoid zombies
            self.scheduled = now()
        super().save(**kwargs)



from setuptools import setup


def readme():
    with open('README.rst', 'r') as f:
        content = f.read()
    return content


setup(name='autotask',
      version='0.1.6',
      description='A django-application for handling asynchronous tasks.',
      long_description=readme(),
      classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
      ],
      keywords='django-application autotask asynchron',
      url='https://bitbucket.org/kbr/autotask',
      author='Klaus Bremer',
      author_email='bremer@bremer-media.de',
      license='MIT',
      packages=[
        'autotask',
        'autotask/management',
        'autotask/management/commands',
        'autotask/migrations',
        'autotask/test',
      ],
      zip_safe=False
)



"""
Function-Decorators for async task-execution by a worker.
"""

import importlib
import pickle
from datetime import timedelta

from django.utils.timezone import now

from .cron import CronScheduler
from .models import (
    WAITING,  # noqa
    RUNNING,  # noqa
    DONE,
    ERROR,
    TaskQueue,
)


class DelayedTask:
    """
    Gives access to a delayed function.
    """

    def __init__(self, pk):
        # pk of TaskQueue item in db
        self.pk = pk

    def _get_task(self):
        try:
            return TaskQueue.objects.get(pk=self.pk)
        except TaskQueue.DoesNotExist:
            return None

    @property
    def ready(self):
        """
        Returns True, False or None.
        In case of None the task is lost and no informations are
        available.
        """
        task = self._get_task()
        if task and not task.is_periodic:
            return task.status == DONE or task.status == ERROR
        return None

    @property
    def status(self):
        """
        Returns the processing status.
        """
        task = self._get_task()
        if task:
            return task.status
        return None

    @property
    def result(self):
        """
        Returns the result of the task.
        """
        task = self._get_task()
        if task:
            try:
                return pickle.loads(task.result)
            except EOFError:
                # if task.result is None this exception will raised
                pass
        return None

    @property
    def error_message(self):
        """
        Returns the error-message as string. If no error has occured
        returns an empty string.
        """
        task = self._get_task()
        if task:
            return task.error_message
        return ''


class DecoratorBase:
    """
    Common functionality for a decorator accepting arguments.
    """

    def __call__(self, function):
        """bind function-label to self.wrapper"""
        self.module_name = function.__module__
        module = importlib.import_module(function.__module__)
        self.function_name = self.template.format(function.__name__)
        setattr(module, self.function_name, function)
        if not self.function_name.endswith('_delayed'):
            # a peridic task will never get called from the application
            # so the wrapper has to be called here:
            self.wrapper()
        return self.wrapper

    def wrapper(self, *args, **kwargs):
        """
        Gets called instead of function. Adds a TaskQueue item for the
        wrapped function and return a DelayedTask objects for accessing
        status-informations and optional results.
        """
        tq = TaskQueue()
        tq.arguments = pickle.dumps((args, kwargs))
        tq.module = self.module_name
        tq.function = self.function_name
        tq = self.configure(tq)
        if tq.is_periodic:
            if self.is_registered(tq):
                return None
        tq.save()
        dt = DelayedTask(tq.pk)
        return dt

    def is_registered(self, tq):
        """
        Returns a boolean whether a task is allready saved in the
        database. This is to prevent periodic tasks registered more than
        once at startup what will happen if the django-project is
        running with more than one process.
        """
        try:
            TaskQueue.objects.get(
                module=tq.module,
                function=tq.function,
                is_periodic=tq.is_periodic)
        except TaskQueue.DoesNotExist:
            return False
        return True


class delayed_task(DecoratorBase):  # noqa
    """
    Decorator to wrap a function for delayed execution by a separate
    worker process.
    :delay: worker waits n seconds before processing
    :retries: on error try to rerun the tasks n times
    :ttl: time to live: after processing the task will stay at least n
    seconds in the database i.e. for accessing the result.

        @delayed_task(optional arguments)
        def long_runner(*args, **kwargs)
        ...
        dt = long_runner()

    The returned object dt is of type DelayedTask
    """
    def __init__(self, delay=0, retries=0, ttl=300):
        self.ttl = timedelta(seconds=ttl)
        self.delay = timedelta(seconds=delay)
        self.retries = retries
        self.template = '{}_delayed'

    def configure(self, tq):
        tq.scheduled = now() + self.delay
        tq.retries = self.retries
        tq.ttl = self.ttl
        tq.is_periodic = False
        return tq


class periodic_task(DecoratorBase):  # noqa
    """
    Decorator for a periodic task running every given number of seconds
    Returns a DelayedTask object. Usage:

        @periodic_task()
        def some_function():
            ...

    Default period is 3600 seconds. If start_now is True the task will
    run as soon as possible and then periodically. Id start_now is False
    the task will be delayed by the given period before running
    periodically.
    """
    def __init__(self, seconds=3600, start_now=False):
        self.timedelta = timedelta(seconds=seconds)
        self.delay = timedelta() if start_now else self.timedelta
        self.template = '{}_periodic'

    def configure(self, tq):
        tq.scheduled = now() + self.delay
        tq.timedelta = self.timedelta
        tq.is_periodic = True
        return tq


class cron_task(DecoratorBase):  # noqa
    """
    Decorator for a periodic task running by a crontab-like schedule:

        @cron_task(optional arguments)
        def the_task()

    The arguments are:

    minutes: list of integers in the range 0-59 or None. [5, 15] will
    run 5 and 15 minutes after full hour. If None the task will run
    every minute.

    hours: list of integers in the range 0-23 or None. [5, 15] will run
    at 5 a.m. and 3 p.m. If None the task will run every hour.

    dow: list of integers in the range 0-6 or None. Describes the days
    of week when a task should run. Monday is 0 and Sunday is 6. [2, 4]
    will run every Wednesday and Friday. If None the task will run every
    day.

    month: list of integers in the range 1-12 or None. Describes the
    month of a year when a task should run. [4, 7] will run in April and
    July. If None the task will run every month.

    dom: list of integers in the range 1-31 or None. Describes the days
    of a month when a task should run. [4, 15] will run at the 4th and
    15th of every month. If None the task will run every day.


    Some examples:

    @cron_task(minutes=[15, 30], hours=[7, 20])
    Runs every day at 7:15, 7:30, 20:15, 20:30

    @cron_task(minutes=[30], hours=[7], dow=[0, 2])
    Runs every Monday and Wednesday at 7:30

    @cron_task(minutes=[30], hours=[7], dow=[0, 2], month=[4, 7])
    Runs every Monday and Wednesday at 7:30 in April and July

    @cron_task(minutes=[30], hours=[7], dom=[1, 15])
    Runs every 1st and 15th of a month at 7:30

    @cron_task(minutes=[30], hours=[7], dow=[0], dom=[1, 15])
    Runs every 1st and 15th of a month and at Mondays at 7:30

    @cron_task(minutes=[30], hours=[7], dow=[0], month=[4], dom=[1, 15])
    Runs at 1st and 15th and at Mondays at 7:30 but only in April

    If all arguments are None the task will run every minute, the same
    as @periodic_task(seconds=60).

    Instead of the separate arguments also a crontab with five patterns
    can be given (pattern order: minutes, hours, day of month, months,
    day of week):
    * * * * *       runs every minute (same as @periodic_task(seconds=60))
    30 7 * * 0,2    runs at 7:30 every Monday and Wednesday.

    """
    def __init__(self, minutes=None, hours=None,
                 dow=None, months=None, dom=None,
                 crontab=None):
        self.template = '{}_cron'
        self.cron_data = {
            'minutes': minutes,
            'hours': hours,
            'dow': dow,
            'months': months,
            'dom': dom,
            'crontab': crontab
        }

    def configure(self, tq):
        cs = CronScheduler(**self.cron_data)
        tq.scheduled = cs.get_next_schedule()
        tq.cron_data = pickle.dumps(self.cron_data)
        tq.is_periodic = True
        return tq



from django.contrib import admin
from .models import TaskQueue


class TaskQueueAdmin(admin.ModelAdmin):
    list_display = ('module', 'function', 'is_periodic', 'scheduled', 'status')


admin.site.register(TaskQueue, TaskQueueAdmin)




import datetime
import importlib
import pickle
import signal
import sys
import threading

from django.conf import settings
from django.db import transaction
from django.utils.timezone import now

from .cron import CronScheduler
from .models import (
    WAITING,
    RUNNING,
    DONE,
    ERROR,
    TaskQueue,
)

# defaults (sec.): can be overridden in settings
AUTOTASK_CLEAN_INTERVALL = 600
AUTOTASK_HANDLE_TASK_IDLE_TIME = 10
AUTOTASK_RETRY_DELAY = 2

# Postgresql exclusive lock
EXCLUSIVE = 'ACCESS EXCLUSIVE'


class TaskHandler:
    """The worker thread for handling callables."""

    def __init__(self):
        self.lock_cmd = 'LOCK TABLE {} IN {} MODE'.format(
            TaskQueue._meta.db_table,
            EXCLUSIVE
        )
        try:
            self.idle_time = settings.AUTOTASK_HANDLE_TASK_IDLE_TIME
        except AttributeError:
            self.idle_time = AUTOTASK_HANDLE_TASK_IDLE_TIME
        try:
            delay = settings.AUTOTASK_RETRY_DELAY
        except AttributeError:
            delay = AUTOTASK_RETRY_DELAY
        self.retry_delay = datetime.timedelta(seconds=delay)

    def __call__(self, exit_event):
        """Entry point for thread start and main loop for worker."""
        task = None
        while True:
            if task:
                self.handle_task(task)
            if exit_event.is_set():
                break
            task = self.get_next_task()
            if task:
                continue
            if exit_event.wait(timeout=self.idle_time):
                break

    def get_next_task(self):
        """
        Returns the next task from the queue on a first come first serve
        basis. Returns None if there is no pending task in the queue.

        This method is postgresql specific because of the locking.
        Locking is done to prevent another process reading task-objects
        in waiting state while this method has already made such a
        query but did not changed the status to waiting. Without such
        a lock it may happen that a task gets handled more than once.
        """
        with transaction.atomic():
            TaskQueue.objects.raw(self.lock_cmd)
            qs = TaskQueue.objects.filter(status=WAITING)
            qs = qs.filter(scheduled__lte=now())
            qs = qs.order_by('scheduled')
            task = qs.first()
            if task:
                task.status = RUNNING
                task.save()
        return task

    def handle_task(self, task):
        """
        Run a delayed or periodic task.
        """
        try:
            task = self._execute(task)
        except Exception as err:
            # catch everything, because it is unknown
            # what may had happen with the callable
            task.error = str(err)
            task.status = ERROR
            if task.is_periodic:
                task.scheduled = self.calculate_schedule(task)
            elif task.retries > 0:
                task.retries -= 1
                task.scheduled = now() + self.retry_delay
                task.status = WAITING
            else:
                # not scheduled again:
                task.expire = now() + task.ttl
        else:
            task.error = ''  # empty: no error
            if task.is_periodic:
                task.status = WAITING
                task.scheduled = self.calculate_schedule(task)
            else:
                task.status = DONE
                task.expire = now() + task.ttl
        task.save()

    def calculate_schedule(self, task):
        """
        Returns the next schedule for a repeating task.
        If task.timedelta is set it is a periodic task, otherwise it is
        a cron task and the next schedule as to be calculated after every
        run.
        """
        if task.function.endswith('_cron'):
            try:
                cron_data = pickle.loads(task.cron_data)
            except Exception as err:
                # don't break on any error.
                # report the error and stay in error-state
                # without a new schedule.
                task.error = str(err)
                task.status = ERROR
                return task.scheduled
            cs = CronScheduler(last_schedule=task.scheduled, **cron_data)
            next_schedule = cs.get_next_schedule()
        else:
            next_schedule = task.scheduled + task.timedelta
        return next_schedule

    def _execute(self, task):
        """
        Find callable, call it and store the result.
        """
        module = importlib.import_module(task.module)
        callable = getattr(module, task.function)
        args, kwargs = pickle.loads(task.arguments)
        result = callable(*args, **kwargs)
        task.result = pickle.dumps(result)
        return task


def run_queue_cleaner(exit_event=None, idle_time=None):
    """
    Clean tasks from the queue with the status DONE and without errors.
    """
    if not idle_time:
        try:
            idle_time = settings.AUTOTASK_CLEAN_INTERVALL
        except AttributeError:
            idle_time = AUTOTASK_CLEAN_INTERVALL
    while True:
        clean_queue()
        if exit_event.wait(timeout=idle_time):
            break
    sys.exit()


def clean_queue():
    """
    Cleanup no longer used task-entries in the database.
    """
    with transaction.atomic():
        qs = TaskQueue.objects.filter(is_periodic=False, expire__lt=now())
        if qs.count():
            qs.delete()


class ShutdownHandler:
    """
    Sets the event for terminating the threads.
    """
    def __init__(self, exit_event):
        self.exit_event = exit_event

    def __call__(self, *args, **kwargs):
        self.exit_event.set()


def start_workers():
    """
    Entry-Point to start the worker in a separate thread and the
    queue-cleaner in the main thread.
    """
    exit_event = threading.Event()
    thread = threading.Thread(target=TaskHandler(), args=(exit_event,))
    thread.start()
    handler = ShutdownHandler(exit_event)
    signal.signal(signal.SIGINT, handler)
    signal.signal(signal.SIGHUP, handler)
    run_queue_cleaner(exit_event)
